"""
   Python SDK for OpenFGA

   API version: 1.x
   Website: https://openfga.dev
   Documentation: https://openfga.dev/docs
   Support: https://openfga.dev/community
   License: [Apache-2.0](https://github.com/openfga/python-sdk/blob/main/LICENSE)

   NOTE: This file was auto generated by OpenAPI Generator (https://openapi-generator.tech). DO NOT EDIT.
"""

import copy
from unittest.mock import Mock

import pytest

from openfga_sdk.configuration import Configuration, RetryParams
from openfga_sdk.exceptions import ApiValueError, FgaValidationException


@pytest.fixture
def configuration():
    return Configuration()


@pytest.fixture
def servers():
    return [
        {"url": "https://example1.fga.example", "description": "Example 1"},
        {"url": "https://example2.fga.example", "description": "Example 2"},
    ]


@pytest.mark.parametrize(
    "scheme,expected",
    [
        ("https", True),
        ("http", True),
        ("ftp", False),
    ],
)
def test_configuration_api_scheme(configuration, scheme, expected):
    assert configuration.api_scheme == "https"

    if expected is True:
        configuration.api_scheme = scheme
        assert configuration.api_scheme == scheme
    else:
        with pytest.raises(FgaValidationException):
            configuration.api_scheme = scheme


def test_configuration_api_host(configuration):
    assert configuration.api_host is None
    configuration.api_host = "fga.example"
    assert configuration.api_host == "fga.example"

    with pytest.raises(ApiValueError):
        configuration.api_host = "ftp://fga.example"
        configuration.is_valid()


def test_configuration_api_url(configuration):
    assert configuration.api_url is None
    configuration.api_url = "https://fga.example/api"
    assert configuration.api_url == "https://fga.example/api"


def test_configuration_store_id(configuration):
    assert configuration.store_id is None
    configuration.store_id = "store123"
    assert configuration.store_id == "store123"


def test_configuration_credentials(configuration):
    assert configuration.credentials is None
    credentials_mock = Mock()
    configuration.credentials = credentials_mock
    assert configuration.credentials == credentials_mock


def test_configuration_retry_params_default_values(configuration):
    assert configuration.retry_params.max_retry == 15
    assert configuration.retry_params.min_wait_in_ms == 100


def test_configuration_retry_params_custom_values(configuration):
    retry_params = RetryParams(max_retry=10, min_wait_in_ms=50)
    configuration.retry_params = retry_params
    assert configuration.retry_params.max_retry == 10
    assert configuration.retry_params.min_wait_in_ms == 50


def test_configuration_retry_params_invalid_max_retry(configuration):
    with pytest.raises(FgaValidationException):
        configuration.retry_params.max_retry = -1


def test_configuration_retry_params_max_retry_greater_than_max(configuration):
    with pytest.raises(FgaValidationException):
        configuration.retry_params.max_retry = 20


def test_configuration_retry_params_invalid_min_wait_in_ms(configuration):
    with pytest.raises(FgaValidationException):
        configuration.retry_params.min_wait_in_ms = -1


class TestConfigurationSetDefaultAndGetDefaultCopy:
    def test_configuration_set_default(self, configuration):
        default_config = Configuration()
        default_config.api_scheme = "https"
        default_config.api_host = "fga.example"
        default_config.store_id = "store123"
        default_config.credentials = Mock(
            _client_id="client123",
            _client_secret="secret123",
            _api_audience="audience123",
            _api_issuer="issuer123",
            _api_token="token123",
        )
        default_config.retry_params = RetryParams(max_retry=10, min_wait_in_ms=50)
        default_config.api_key = {"api_key1": "key1", "api_key2": "key2"}
        default_config.api_key_prefix = {"api_key1": "Bearer", "api_key2": "Bearer"}
        default_config.username = "user"
        default_config.password = "pass"
        default_config.discard_unknown_keys = True
        default_config.disabled_client_side_validations = "maxLength,minLength"
        default_config.server_index = 1
        default_config.server_variables = {"variable1": "value1", "variable2": "value2"}
        default_config.server_operation_index = {"operation1": 1, "operation2": 2}
        default_config.server_operation_variables = {
            "operation1": {"var1": "val1"},
            "operation2": {"var2": "val2"},
        }
        default_config.ssl_ca_cert = "/path/to/ca_cert.pem"
        default_config.api_url = "https://fga.example/api"
        Configuration.set_default(default_config)

        assert Configuration._default.api_scheme == "https"
        assert Configuration._default.api_host == "fga.example"
        assert Configuration._default.store_id == "store123"
        assert (
            isinstance(Configuration._default.credentials, Mock)
            or Configuration._default.credentials is None
        )
        if Configuration._default.credentials:
            assert Configuration._default.credentials._client_id == "client123"
            assert Configuration._default.credentials._client_secret == "secret123"
            assert Configuration._default.credentials._api_audience == "audience123"
            assert Configuration._default.credentials._api_issuer == "issuer123"
            assert Configuration._default.credentials._api_token == "token123"
        assert Configuration._default.retry_params.max_retry == 10
        assert Configuration._default.retry_params.min_wait_in_ms == 50
        assert Configuration._default.api_key == {
            "api_key1": "key1",
            "api_key2": "key2",
        }
        assert Configuration._default.api_key_prefix == {
            "api_key1": "Bearer",
            "api_key2": "Bearer",
        }
        assert Configuration._default.username == "user"
        assert Configuration._default.password == "pass"
        assert Configuration._default.discard_unknown_keys is True
        assert Configuration._default.disabled_client_side_validations == {
            "maxLength",
            "minLength",
        }
        assert Configuration._default.server_index == 1
        assert Configuration._default.server_variables == {
            "variable1": "value1",
            "variable2": "value2",
        }
        assert Configuration._default.server_operation_index == {
            "operation1": 1,
            "operation2": 2,
        }
        assert Configuration._default.server_operation_variables == {
            "operation1": {"var1": "val1"},
            "operation2": {"var2": "val2"},
        }
        assert Configuration._default.ssl_ca_cert == "/path/to/ca_cert.pem"
        assert Configuration._default.api_url == "https://fga.example/api"

    def test_configuration_get_default_copy(self, configuration):
        default_config = Configuration()
        default_config.api_scheme = "https"
        default_config.api_host = "fga.example"
        default_config.store_id = "store123"
        default_config.credentials = Mock(
            _client_id="client123",
            _client_secret="secret123",
            _api_audience="audience123",
            _api_issuer="issuer123",
            _api_token="token123",
        )
        default_config.retry_params = RetryParams(max_retry=10, min_wait_in_ms=50)
        default_config.api_key = {"api_key1": "key1", "api_key2": "key2"}
        default_config.api_key_prefix = {"api_key1": "Bearer", "api_key2": "Bearer"}
        default_config.username = "user"
        default_config.password = "pass"
        default_config.discard_unknown_keys = True
        default_config.disabled_client_side_validations = "maxLength,minLength"
        default_config.server_index = 1
        default_config.server_variables = {"variable1": "value1", "variable2": "value2"}
        default_config.server_operation_index = {"operation1": 1, "operation2": 2}
        default_config.server_operation_variables = {
            "operation1": {"var1": "val1"},
            "operation2": {"var2": "val2"},
        }
        default_config.ssl_ca_cert = "/path/to/ca_cert.pem"
        default_config.api_url = "https://fga.example/api"
        Configuration.set_default(default_config)

        copied_config = Configuration.get_default_copy()

        assert copied_config.api_scheme == "https"
        assert copied_config.api_host == "fga.example"
        assert copied_config.store_id == "store123"
        assert (
            isinstance(copied_config.credentials, Mock)
            or copied_config.credentials is None
        )
        if copied_config.credentials:
            assert copied_config.credentials._client_id == "client123"
            assert copied_config.credentials._client_secret == "secret123"
            assert copied_config.credentials._api_audience == "audience123"
            assert copied_config.credentials._api_issuer == "issuer123"
            assert copied_config.credentials._api_token == "token123"


class TestConfigurationValidityChecks:
    def test_configuration_is_valid_missing_api_url(self, configuration):
        with pytest.raises(FgaValidationException):
            configuration.is_valid()

    def test_configuration_is_valid_invalid_store_id(self, configuration):
        configuration.api_url = "https://fga.example"
        configuration.store_id = "invalid_ulid"
        with pytest.raises(FgaValidationException):
            configuration.is_valid()

    def test_configuration_is_valid(self, configuration):
        configuration.api_url = "https://fga.example"
        configuration.store_id = "01F9ZCDXDZBXK83WVMY1VZT23V"
        assert configuration.is_valid() is None


class TestConfigurationLogging:
    def test_configuration_logger_format(self, configuration):
        assert configuration.logger_format == "%(asctime)s %(levelname)s %(message)s"
        configuration.logger_format = "%(levelname)s: %(message)s"
        assert configuration.logger_format == "%(levelname)s: %(message)s"

    def test_configuration_debug(self, configuration):
        assert not configuration.debug
        configuration.debug = True
        assert configuration.debug
        configuration.debug = False
        assert not configuration.debug

    def test_configuration_logger_file(self, configuration):
        assert configuration.logger_file is None
        configuration.logger_file = "debug.log"
        assert configuration.logger_file == "debug.log"

    def test_configuration_disabled_client_side_validations(self, configuration):
        assert configuration.disabled_client_side_validations == {}
        configuration.disabled_client_side_validations = "maxLength,minLength"
        assert configuration.disabled_client_side_validations == {
            "maxLength",
            "minLength",
        }

    def test_configuration_disabled_client_side_validations_invalid_keyword(
        self, configuration
    ):
        with pytest.raises(FgaValidationException):
            configuration.disabled_client_side_validations = "invalidKeyword"


class TestConfigurationHostSettings:
    def test_configuration_get_host_settings(self, configuration):
        assert configuration.get_host_settings() == [
            {"url": "", "description": "No description provided"}
        ]

    def test_configuration_get_host_from_settings(self, configuration, servers):
        assert (
            configuration.get_host_from_settings(0, servers=servers)
            == "https://example1.fga.example"
        )
        assert (
            configuration.get_host_from_settings(1, servers=servers)
            == "https://example2.fga.example"
        )

    def test_configuration_get_host_from_settings_invalid_value(self, configuration):
        servers = [
            {
                "url": "https://{region}.fga.example",
                "description": "Example",
                "variables": {"region": {"default_value": "us"}},
            }
        ]
        variables = {"region": "invalid"}
        with pytest.raises(ValueError):
            configuration.get_host_from_settings(999, variables={"var": "value"})


class TestConfigurationMiscellaneous:
    def test_configuration_get_api_key_with_prefix(self, configuration):
        configuration.api_key = {"api_key": "123"}
        configuration.api_key_prefix = {"api_key": "Bearer"}
        assert configuration.get_api_key_with_prefix("api_key") == "Bearer 123"

    def test_configuration_get_basic_auth_token(self, configuration):
        configuration.username = "user"
        configuration.password = "pass"
        assert configuration.get_basic_auth_token() == "Basic dXNlcjpwYXNz"

    def test_configuration_auth_settings(self, configuration):
        assert configuration.auth_settings() == {}

    def test_configuration_to_debug_report(self, configuration):
        report = configuration.to_debug_report()
        assert "Python SDK Debug Report" in report
        assert "OS" in report
        assert "Python Version" in report
        assert "Version of the API" in report
        assert "SDK Package Version" in report

    def test_configuration_deepcopy(self, configuration):
        # Create a Configuration object with some values
        config = Configuration(
            api_scheme="https",
            api_host="fga.example",
            store_id="store123",
            credentials=Mock(
                _client_id="client123",
                _client_secret="secret123",
                _api_audience="audience123",
                _api_issuer="issuer123",
                _api_token="token123",
            ),
            retry_params=RetryParams(max_retry=10, min_wait_in_ms=50),
            api_key={"api_key1": "key1", "api_key2": "key2"},
            api_key_prefix={"api_key1": "Bearer", "api_key2": "Bearer"},
            username="user",
            password="pass",
            discard_unknown_keys=True,
            disabled_client_side_validations="maxLength,minLength",
            server_index=1,
            server_variables={"variable1": "value1", "variable2": "value2"},
            server_operation_index={"operation1": 1, "operation2": 2},
            server_operation_variables={
                "operation1": {"var1": "val1"},
                "operation2": {"var2": "val2"},
            },
            ssl_ca_cert="/path/to/ca_cert.pem",
            api_url="https://fga.example/api",
        )

        # Perform deep copy
        copied_config = copy.deepcopy(config)

        # Verify all attributes of copied object are equal to original object
        assert copied_config.api_scheme == config.api_scheme
        assert copied_config.api_host == config.api_host
        assert copied_config.store_id == config.store_id
        assert (
            isinstance(Configuration._default.credentials, Mock)
            or Configuration._default.credentials is None
        )
        assert Configuration._default.retry_params.max_retry == 10
        assert Configuration._default.retry_params.min_wait_in_ms == 50
        assert copied_config.api_key == config.api_key
        assert copied_config.api_key_prefix == config.api_key_prefix
        assert copied_config.username == config.username
        assert copied_config.password == config.password
        assert copied_config.discard_unknown_keys == config.discard_unknown_keys
        assert (
            copied_config.disabled_client_side_validations
            == config.disabled_client_side_validations
        )
        assert copied_config.server_index == config.server_index
        assert copied_config.server_variables == config.server_variables
        assert copied_config.server_operation_index == config.server_operation_index
        assert (
            copied_config.server_operation_variables
            == config.server_operation_variables
        )
        assert copied_config.ssl_ca_cert == config.ssl_ca_cert
        assert copied_config.api_url == config.api_url
